import collections

import tensorflow as tf
import numpy as np
import math as m

from Neural_Networks import vae_utils

# based on implementation here:
# https://github.com/tensorflow/models/blob/master/autoencoder/autoencoder_models/VariationalAutoencoder.py

SMALL_CONSTANT = 1e-6

class VariationalAutoencoder(object):

    def __init__(self, name, n_input, n_hidden, n_weights, middle="gaussian"):
        
        self.n_input = n_input
        self.n_hidden = n_hidden
        self.n_weights = n_weights
        self.name = name
        self.middle = middle
        self.bias_start = 0.0

        network_weights = self._create_weights()
        self.weights = network_weights

        self.nonlinearity = tf.nn.relu
        self.nonlinearity2 = tf.nn.leaky_relu


    def _calc_z_mean_and_sigma(self,x):
        with tf.name_scope("VICI_VAE_encoder"):
            
            layer_1 = tf.add(tf.matmul(x, self.weights['VICI_VAE_encoder']['W_layer_1']), self.weights['VICI_VAE_encoder']['b_layer_1'])
            previous_layer_out = self.nonlinearity(layer_1)

            layer_2 = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_VAE_encoder']['W_layer_2']), self.weights['VICI_VAE_encoder']['b_layer_2'])
            previous_layer_out = self.nonlinearity(layer_2)

            z_mean = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_VAE_encoder']['W_layer_out_mu']), self.weights['VICI_VAE_encoder']['b_layer_out_mu'])
            z_log_sigma_sq = tf.add(tf.matmul(previous_layer_out, self.weights['VICI_VAE_encoder']['W_layer_out_log_sigma']), self.weights['VICI_VAE_encoder']['b_layer_out_log_sigma'])

            tf.summary.histogram("z_mean", z_mean)
            tf.summary.histogram("z_log_sigma_sq", z_log_sigma_sq)

            return z_mean, z_log_sigma_sq

    def _sample_from_gaussian_dist(self, num_rows, num_cols, mean, log_sigma_sq):
        with tf.name_scope("sample_in_z_space"):
            eps = tf.random_normal([num_rows, num_cols], 0, 1., dtype=tf.float32)
            sample = tf.add(mean, tf.multiply(tf.sqrt(tf.exp(log_sigma_sq)), eps))
        return sample

    def _create_weights(self):
        all_weights = collections.OrderedDict()
        with tf.variable_scope("VICI_VAE_ENC"):
            # Encoder
            all_weights['VICI_VAE_encoder'] = collections.OrderedDict()
            hidden_number_encoder = self.n_weights

            all_weights['VICI_VAE_encoder']['W_layer_1'] = tf.Variable(vae_utils.xavier_init(self.n_input, hidden_number_encoder), dtype=tf.float32)
            all_weights['VICI_VAE_encoder']['b_layer_1'] = tf.Variable(tf.zeros([hidden_number_encoder], dtype=tf.float32) * self.bias_start)
            tf.summary.histogram("W_layer_1", all_weights['VICI_VAE_encoder']['W_layer_1'])
    
            all_weights['VICI_VAE_encoder']['W_layer_2'] = tf.Variable(vae_utils.xavier_init(hidden_number_encoder, hidden_number_encoder), dtype=tf.float32)
            all_weights['VICI_VAE_encoder']['b_layer_2'] = tf.Variable(tf.zeros([hidden_number_encoder], dtype=tf.float32) * self.bias_start)
            tf.summary.histogram("W_layer_2", all_weights['VICI_VAE_encoder']['W_layer_2'])
    
            all_weights['VICI_VAE_encoder']['W_layer_out_mu'] = tf.Variable(vae_utils.xavier_init(hidden_number_encoder, self.n_hidden),dtype=tf.float32)
            all_weights['VICI_VAE_encoder']['b_layer_out_mu'] = tf.Variable(tf.zeros([self.n_hidden], dtype=tf.float32) * self.bias_start, dtype=tf.float32)
            tf.summary.histogram("W_layer_out_mu", all_weights['VICI_VAE_encoder']['W_layer_out_mu'])
    
            all_weights['VICI_VAE_encoder']['W_layer_out_log_sigma'] = tf.Variable(vae_utils.xavier_init(hidden_number_encoder, self.n_hidden), dtype=tf.float32)
            all_weights['VICI_VAE_encoder']['b_layer_out_log_sigma'] = tf.Variable(tf.zeros([self.n_hidden], dtype=tf.float32) * self.bias_start, dtype=tf.float32)
            tf.summary.histogram("W_layer_out_log_sigma", all_weights['VICI_VAE_encoder']['W_layer_out_log_sigma'])
            
            all_weights['prior_param'] = collections.OrderedDict()
        
        return all_weights
